# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Package management parsing and handling logic
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Factored from specialized handlers in cases that created import loops.

:author: Itamar Ostricher
"""


import requests


LAUNCHPAD_URL = ('https://launchpad.net/api/1.0/'
                 '~{ppa_owner}/+archive/{ppa_name}')
LAUNCHPAD_SOURCE_LINE = ('deb http://ppa.launchpad.net/{ppa_owner}/{ppa_name}/'
                         '{distro_id} {distro_codename} main')
VALID_SOURCE_TYPES = frozenset(('deb',))  # 'deb-src'


def format_apt_specifier(target):
    if target.props.version:
        return '{0.package}={0.version}'.format(target.props)
    return '{0.package}'.format(target.props)


def expand_ppa(path: str, distro: dict):
    ppa = path.split(':')[1].split('/')
    ppa_owner = ppa[0]
    ppa_name = ppa[1] if len(ppa) > 0 else 'ppa'
    source_line = LAUNCHPAD_SOURCE_LINE.format(
        ppa_owner=ppa_owner, ppa_name=ppa_name,
        distro_id=distro.get('id', 'ubuntu').lower(),
        distro_codename=distro.get('codename', 'trusty'))
    return source_line, ppa_owner, ppa_name


def parse_apt_repository(build_context, target, distro):
    source_line = target.props.source
    apt_key_cmd = None
    # Parse PPA
    if source_line.startswith('ppa:'):
        source_line, ppa_owner, ppa_name = expand_ppa(source_line, distro)
        response = requests.get(
            LAUNCHPAD_URL.format(ppa_owner=ppa_owner, ppa_name=ppa_name),
            headers={'Accept': 'application/json'})
        if response.status_code != 200:
            raise RuntimeError('Failed getting PPA info for {}'.format(target))
        target.props.key = response.json()['signing_key_fingerprint']
    # Build apt-key command
    if target.props.key:
        apt_key_cmd = ('apt-key adv --keyserver {keyserver} --recv {key}'
                       .format(**target.props))
    # Clean up and validate apt source line
    chunks = source_line.split('#', 1)[0].strip().split()
    if not chunks or chunks[0] not in VALID_SOURCE_TYPES:
        raise ValueError('Invalid source line "{}"'.format(source_line))
    source_line = ' '.join(chunks)
    return source_line, apt_key_cmd


def format_pypi_specifier(target):
    if target.props.version:
        return '{0.package}=={0.version}'.format(target.props)
    return '{0.package}'.format(target.props)



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Docker module
~~~~~~~~~~~~~~~~~~

NOT to be confused with the Docker builder...

:author: Itamar Ostricher
"""


from collections import defaultdict, deque
import os
from os.path import (
    abspath, basename, dirname, isfile, join, relpath, samefile, split)
import platform
import shutil

from ostrich.utils.path import commonpath
from ostrich.utils.proc import run, PIPE, CalledProcessError
from ostrich.utils.text import get_safe_path
from ostrich.utils.collections import listify

from .config import Config
from .logging import make_logger
from .builders.nodejs import format_npm_specifier
from .builders.ruby import format_gem_specifier
from .pkgmgmt import (
    format_apt_specifier, format_pypi_specifier, parse_apt_repository)
from . import target_utils
from .utils import link_artifacts, link_node, rmtree, yprint


logger = make_logger(__name__)


def make_pip_requirements(pip_requirements: set, pip_req_file_path: str):
    if pip_requirements:
        try:
            with open(pip_req_file_path, 'r') as pip_req_file:
                if (set(line.strip() for line in pip_req_file) ==
                        pip_requirements):
                    # no change in requirements file - don't rewrite it
                    # (it appears that rewriting may trigger spurious rebuilds)
                    logger.debug(
                        'Short circuiting requirements file generation')
                    return True
        except FileNotFoundError:
            pass
        with open(pip_req_file_path, 'w') as pip_req_file:
            pip_req_file.write('\n'.join(sorted(pip_requirements)) + '\n')
        return True
    elif isfile(pip_req_file_path):
        # delete remnant requirements file
        os.remove(pip_req_file_path)
        return False


def make_apt_sources_list(apt_sources: list, apt_sources_file_path: str):
    if apt_sources:
        try:
            with open(apt_sources_file_path, 'r') as apt_sources_file:
                if (set(line.strip() for line in apt_sources_file) ==
                        set(apt_sources)):
                    # no change in sources file - don't rewrite it
                    logger.debug(
                        'Short circuiting sources list file generation')
                    return True
        except FileNotFoundError:
            pass
        with open(apt_sources_file_path, 'w') as apt_sources_file:
            apt_sources_file.write('\n'.join(sorted(apt_sources)) + '\n')
        return True
    elif isfile(apt_sources_file_path):
        # delete remnant sources list file
        os.remove(apt_sources_file_path)
        return False


def get_image_name(target):
    # left-stripping ":" to remove the build-module separator for root images,
    # since Docker image names must begin with an alphanumeric character
    return (target.props.image_name if target.props.image_name
            else get_safe_path(target.name.lstrip(':')))


def get_remote_image_name(name: str, tag: str, image_caching_behavior: dict):
    remote_image_name = image_caching_behavior.get('remote_image_name', name)
    remote_image_tag = image_caching_behavior.get('remote_image_tag', tag)
    return '{}:{}'.format(remote_image_name, remote_image_tag)


def format_qualified_image_name(target):
    if target.builder_name == 'ExtDockerImage':
        if target.props.tag:
            return '{}:{}'.format(target.props.image, target.props.tag)
        return target.props.image
    elif 'docker_image_id' in target.props:
        return target.props.docker_image_id
    elif target.builder_name == 'DockerImage':
        return '{}:{}'.format(get_image_name(target), target.props.image_tag)
    else:
        raise TypeError(target)


def get_cached_image_id(qualified_image_name):
    docker_cmd = ['docker', 'images', '-q', qualified_image_name]
    result = run(docker_cmd, stdout=PIPE)
    if result.stdout:
        return result.stdout.decode('utf8').strip()
    return None


def pull_docker_image(qualified_image_name: str, pull_cmd: list):
    docker_pull_cmd = pull_cmd + [qualified_image_name]
    logger.debug('Pulling Docker image {} using command {}',
                 qualified_image_name, docker_pull_cmd)
    run(docker_pull_cmd, check=True)


def push_docker_image(qualified_image_name: str, push_cmd: list):
    docker_push_cmd = push_cmd + [qualified_image_name]
    logger.debug('Pushing Docker image {} using command {}',
                 qualified_image_name, docker_push_cmd)
    run(docker_push_cmd, check=True)


def tag_docker_image(src_image, tag_as_image):
    if src_image == tag_as_image:
        logger.debug('Skipping Docker tag for identical src and dest {}',
                     src_image)
        return
    docker_tag_cmd = ['docker', 'tag', src_image, tag_as_image]
    logger.debug('Tagging Docker image {} as {} using command {}',
                 src_image, tag_as_image, docker_tag_cmd)
    run(docker_tag_cmd, check=True)


def handle_build_cache(
        build_context, name: str, tag: str, image_caching_behavior: dict):
    """Handle Docker image build cache.

    Return image ID if image is cached, and there's no need to redo the build.
    Return None if need to build the image (whether cahced locally or not).
    Raise RuntimeError if not allowed to build the image because of state of
    local cache.

    TODO(itamar): figure out a better name for this function, that reflects
    what it returns (e.g. `get_cached_image_id`),
    without "surprising" the caller with the potential of long
    and non-trivial operations that are not usually expected from functions
    with such names.
    """
    local_image = '{}:{}'.format(name, tag)
    remote_image = get_remote_image_name(name, tag, image_caching_behavior)
    pull_if_not_cached = image_caching_behavior.get(
        'pull_if_not_cached', False)
    pull_if_cached = image_caching_behavior.get(
        'pull_if_cached', False)
    allow_build_if_not_cached = image_caching_behavior.get(
        'allow_build_if_not_cached', True)
    skip_build_if_cached = image_caching_behavior.get(
        'skip_build_if_cached', False)
    if pull_if_cached or (pull_if_not_cached and
                          get_cached_image_id(remote_image) is None):
        try:
            pull_docker_image(remote_image, build_context.conf.docker_pull_cmd)
        except CalledProcessError:
            pass
    local_image = '{}:{}'.format(name, tag)
    if skip_build_if_cached and get_cached_image_id(remote_image) is not None:
        tag_docker_image(remote_image, local_image)
        return get_cached_image_id(local_image)
    if ((not allow_build_if_not_cached) and
            get_cached_image_id(remote_image) is None):
        raise RuntimeError('No cached image for {}'.format(local_image))
    return None


def build_docker_image(
        build_context, name: str, tag: str, base_image, deps: list=None,
        env: dict=None, work_dir: str=None, truncate_common_parent: str=None,
        entrypoint: list=None, cmd: list=None, full_path_cmd: bool=False,
        distro: dict=None, image_caching_behavior: dict=None,
        runtime_params: dict=None, ybt_bin_path: str=None,
        build_user: str=None, run_user: str=None, labels: dict=None,
        no_artifacts: bool=False):
    """Build Docker image, and return a (image_id, image_name:tag) tuple of
       built image, if built successfully.

    Notes:
    Using the given image name & tag as they are, but using the global host
    Docker image namespace (as opposed to a private-project-workspace),
    so collisions between projects are possible (and very likely, e.g., when
    used in a CI environment, or shared machine use-case).
    Trying to address this issue to some extent by using the image ID after
    it is built, which is unique.
    There's a race condition between "build" and "get ID" - ignoring this at
    the moment.
    Also, I'm not sure about Docker's garbage collection...
    If I use the image ID in other places, and someone else "grabbed" my image
    name and tag (so now my image ID is floating), is it still safe to use
    the ID? Or is it going to be garbage collected and cleaned up sometime?
    From my experiments, the "floating image ID" was left alone (usable),
    but prone to "manual cleanups".
    Also ignoring this at the moment...
    Thought about an alternative approach based on first building an image
    with a randomly generated tag, so I can use that safely later, and tag it
    to the requested tag.
    Decided against it, seeing that every run increases the local Docker
    images spam significantly with a bunch of random tags, making it even less
    useful.
    Documenting it here to remember it was considered, and to discuss it
    further in case anyone thinks it's a better idea than what I went with.
    """
    docker_image = '{}:{}'.format(name, tag)
    if image_caching_behavior is None:
        image_caching_behavior = {}
    image_id = handle_build_cache(
        build_context, name, tag, image_caching_behavior)
    if image_id:
        yprint(build_context.conf,
               'Skipping build of cached Docker image', docker_image)
        return {'image_id': image_id}
    # create directory for this target under a private builder workspace
    workspace_dir = build_context.get_workspace('DockerBuilder', docker_image)
    # generate Dockerfile and build it
    dockerfile_path = join(workspace_dir, 'Dockerfile')
    dockerfile = [
        'FROM {}\n'.format(format_qualified_image_name(base_image)),
        'ARG DEBIAN_FRONTEND=noninteractive\n',
    ]
    if build_user:
        dockerfile.append('USER {}\n'.format(build_user))
    all_artifacts = defaultdict(set)
    gen_artifacts = {}
    apt_repo_deps = []
    effective_env = {}
    KNOWN_RUNTIME_PARAMS = frozenset((
        'ports', 'volumes', 'container_name', 'daemonize', 'interactive',
        'term', 'auto_it', 'rm', 'env', 'work_dir', 'impersonate'))
    if runtime_params is None:
        runtime_params = {}
    runtime_params['ports'] = listify(runtime_params.get('ports'))
    runtime_params['volumes'] = listify(runtime_params.get('volumes'))
    runtime_params['env'] = dict(runtime_params.get('env', {}))
    env_manipulations = {}
    packaging_layers = []

    def add_package(pkg_type, pkg_spec):
        """Add package specification of certain package type.

        Uses last layer if matches package type, otherwise opens a new layer.

        This can result "Docker layer framgantation", by opening and closing
        many layers.
        No optimization is performed on detecting opportunities to merge layers
        that were split just because of arbitrary topological sort decision
        (tie breaker), and not a real topology in the target graph.
        Such an optimization could be done here by inspecting the graph
        directly, but decided not to go into it at this stage, since it's not
        clear it's beneficial overall (e.g. better to have more layers so
        some of them can remain cached if others change).
        A better optimization (also not implemented) could be to do topological
        sort tie breaking based on Docker-cache optimization - e.g., move new
        things to new layers in order to keep old things in cached layers.
        """
        if len(packaging_layers) == 0:
            layer = (pkg_type, list())
            packaging_layers.append(layer)
        else:
            layer = packaging_layers[-1]
            if pkg_type != layer[0]:
                layer = (pkg_type, list())
                packaging_layers.append(layer)
        layer[1].append(pkg_spec)

    def check_env_overrides(new_vars: set, op_kind: str, vars_source: str):
        overridden_vars = new_vars.intersection(effective_env.keys())
        if overridden_vars:
            raise ValueError(
                'Following env vars {} from {} override previously set vars '
                'during build of Docker image "{}": {}'.format(
                    op_kind, vars_source, docker_image,
                    ', '.join(overridden_vars)))
        if op_kind == 'set':
            overridden_vars = new_vars.intersection(env_manipulations.keys())
            if overridden_vars:
                raise ValueError(
                    'Following env vars {} from {} override previous '
                    'manipulations during build of Docker image "{}": {}'
                    .format(op_kind, vars_source, docker_image,
                            ', '.join(overridden_vars)))

    def update_runtime_params(new_rt_param: dict, params_source: str):
        invalid_keys = set(
            new_rt_param.keys()).difference(KNOWN_RUNTIME_PARAMS)
        if invalid_keys:
            raise ValueError(
                'Unknown keys in runtime params of {}: {}'.format(
                    params_source, ', '.join(invalid_keys)))
        # TODO(itamar): check for invalid values and inconsistencies
        runtime_params['ports'].extend(listify(new_rt_param.get('ports')))
        runtime_params['volumes'].extend(listify(new_rt_param.get('volumes')))
        runtime_params['env'].update(dict(runtime_params.get('env', {})))
        for param in ('container_name', 'daemonize', 'interactive', 'term',
                      'auto_it', 'rm', 'work_dir', 'impersonate'):
            if param in new_rt_param:
                # TODO(itamar): check conflicting overrides
                runtime_params[param] = new_rt_param[param]

    if deps is None:
        deps = []
    # Get all base image deps, so when building this image we can skip adding
    # deps that already exist in the base image.
    base_image_deps = set(dep.name for dep in
                          build_context.walk_target_graph([base_image.name]))
    for dep in deps:
        if not distro and 'distro' in dep.props:
            distro = dep.props.distro
        if 'runtime_params' in dep.props:
            update_runtime_params(dep.props.runtime_params,
                                  'dependency {}'.format(dep.name))

        if dep.name in base_image_deps:
            logger.debug('Skipping base image dep {}', dep.name)
            continue
        if not no_artifacts:
            for kind, artifacts in dep.artifacts.items():
                if kind == 'gen':
                    gen_artifacts.update(artifacts)
                else:
                    all_artifacts[kind].update(artifacts)

        PACKAGING_PARAMS = frozenset(('set_env', 'semicolon_join_env'))
        invalid_keys = set(
            dep.props.packaging_params.keys()).difference(PACKAGING_PARAMS)
        if invalid_keys:
            raise ValueError(
                'Unknown keys in packaging params of target "{}": {}'.format(
                    dep.name, ', '.join(invalid_keys)))
        if 'set_env' in dep.props.packaging_params:
            dep_env = dep.props.packaging_params['set_env']
            check_env_overrides(
                set(dep_env.keys()), 'set', 'dependency {}'.format(dep.name))
            effective_env.update(dep_env)
        if 'semicolon_join_env' in dep.props.packaging_params:
            append_env = dep.props.packaging_params['semicolon_join_env']
            check_env_overrides(set(append_env.keys()), 'manipulations',
                                'dependency {}'.format(dep.name))
            for key, value in append_env.items():
                env_manip = env_manipulations.setdefault(
                    key, ['${{{}}}'.format(key)])
                if value not in env_manip:
                    env_manip.append(value)

        if 'apt-repository' in dep.tags:
            apt_repo_deps.append(dep)
        if 'apt-installable' in dep.tags:
            add_package('apt', format_apt_specifier(dep))
        if 'pip-installable' in dep.tags:
            add_package(dep.props.pip, format_pypi_specifier(dep))
        if 'custom-installer' in dep.tags:
            add_package('custom', dep.props.installer_desc)
        if 'npm-installable' in dep.tags:
            if dep.props.global_install:
                add_package('npm-global', format_npm_specifier(dep))
            else:
                add_package('npm-local', format_npm_specifier(dep))
        if 'gem-installable' in dep.tags:
            add_package('gem', format_gem_specifier(dep))

    # Add environment variables (one layer)
    if env:
        check_env_overrides(set(env.keys()), 'set', 'the target')
        effective_env.update(env)
    for key, value in env_manipulations.items():
        effective_env[key] = ':'.join(value)
    if effective_env:
        dockerfile.append(
            'ENV {}\n'.format(
                ' '.join('{}="{}"'.format(key, value)
                         for key, value in sorted(effective_env.items()))))

    apt_key_cmds = []
    apt_repositories = []
    for dep in apt_repo_deps:
        source_line, apt_key_cmd = parse_apt_repository(
            build_context, dep, distro)
        apt_repositories.append(source_line)
        if apt_key_cmd:
            apt_key_cmds.append(apt_key_cmd)
    # Handle apt keys (one layer for all)
    if apt_key_cmds:
        dockerfile.append(
            'RUN {}\n'.format(' && '.join(apt_key_cmds)))
    # Handle apt repositories (one layer for all)
    if apt_repositories:
        list_name = '{}.list'.format(name)
        apt_src_file = join(workspace_dir, list_name)
        if make_apt_sources_list(apt_repositories, apt_src_file):
            dockerfile.append(
                'COPY {} /etc/apt/sources.list.d/\n'.format(list_name))

    custom_cnt = 0
    pip_req_cnt = defaultdict(int)

    def install_npm(npm_packages: list, global_install: bool):
        if npm_packages:
            if not global_install:
                dockerfile.append('WORKDIR /usr/src\n')
            dockerfile.append(
                'RUN npm install {} {}\n'.format(
                    ' '.join(npm_packages),
                    '--global' if global_install else '&& npm dedupe'))

    for layer in packaging_layers:
        pkg_type, packages = layer

        if pkg_type == 'apt':
            dockerfile.append(
                'RUN apt-get update -y && apt-get install '
                '--no-install-recommends -y {} '
                '&& rm -rf /var/lib/apt/lists/*\n'.format(
                    ' '.join(sorted(packages))))

        elif pkg_type == 'custom':
            # Handle custom installers (2 layers per occurrence)
            custom_cnt += 1
            packages_dir = 'packages{}'.format(custom_cnt)
            tmp_install = '/tmp/install{}'.format(custom_cnt)
            workspace_packages_dir = join(workspace_dir, packages_dir)
            rmtree(workspace_packages_dir)
            os.makedirs(workspace_packages_dir)
            run_installers = []
            for custom_installer in packages:
                package_tar = basename(custom_installer.package)
                os.link(custom_installer.package,
                        join(workspace_packages_dir, package_tar))
                run_installers.extend([
                    'tar -xf {0}/{1} -C {0}'.format(tmp_install, package_tar),
                    'cd {}/{}'.format(tmp_install, custom_installer.name),
                    './{}'.format(custom_installer.install_script),
                ])
            dockerfile.extend([
                'COPY {} {}\n'.format(packages_dir, tmp_install),
                'RUN {} && cd / && rm -rf {}\n'.format(
                    ' && '.join(run_installers), tmp_install),
            ])

        elif pkg_type.startswith('pip'):
            # Handle pip packages (2 layers per occurrence)
            req_fname = 'requirements_{}_{}.txt'.format(
                pkg_type, pip_req_cnt[pkg_type] + 1)
            pip_req_file = join(workspace_dir, req_fname)
            if make_pip_requirements(packages, pip_req_file):
                upgrade_pip = (
                    '{pip} install --no-cache-dir --upgrade pip && '
                    .format(pip=pkg_type)
                    if pip_req_cnt[pkg_type] == 0 else '')
                dockerfile.extend([
                    'COPY {} /usr/src/\n'.format(req_fname),
                    'RUN {upgrade_pip}'
                    '{pip} install --no-cache-dir -r /usr/src/{reqs}\n'
                    .format(upgrade_pip=upgrade_pip, pip=pkg_type,
                            reqs=req_fname)
                ])
                pip_req_cnt[pkg_type] += 1

        elif pkg_type == 'npm-global':
            # Handle npm global packages (1 layer per occurrence)
            install_npm(packages, True)

        elif pkg_type == 'npm-local':
            # Handle npm local packages (1 layer per occurrence)
            install_npm(packages, False)

        elif pkg_type == 'gem':
            # Handle gem (ruby) packages (1 layer per occurrence)
            dockerfile.append(
                'RUN gem install {}\n'.format(' '.join(packages)))

    if work_dir:
        dockerfile.append('WORKDIR {}\n'.format(work_dir))

    if not no_artifacts:
        # Handle copying data to the image
        # start with removing the workspace src dir, to avoid any spurious and
        # leftover files in there - I think this is better than walking it and
        # looking for files to remove - doesn't seem this would scale well
        workspace_src_dir = join(workspace_dir, 'src')
        rmtree(workspace_src_dir)
        # sync artifacts between project and `workspace_src_dir`
        num_linked = 0
        # link app artifacts (with common parent truncation)
        app_artifacts = all_artifacts.pop('app', None)
        if app_artifacts:
            num_linked += link_artifacts(
                app_artifacts, join(workspace_src_dir, 'app'),
                truncate_common_parent, build_context.conf)
        # link generated artifacts
        for dest, src in gen_artifacts.items():
            link_node(join(build_context.conf.project_root, src),
                      join(workspace_src_dir, 'gen', dest))
            num_linked += 1
        # link "other" artifacts
        for kind, artifacts in all_artifacts.items():
            num_linked += link_artifacts(
                artifacts, join(workspace_src_dir, kind), None,
                build_context.conf)
        if num_linked > 0:
            dockerfile.append('COPY src /usr/src\n')

    if labels:
        for label, value in sorted(labels.items()):
            dockerfile.append('LABEL "{}"="{}"\n'.format(label, value))

    def format_docker_cmd(docker_cmd):
        return ('"{}"'.format(cmd) for cmd in docker_cmd)

    if run_user:
        dockerfile.append('USER {}\n'.format(run_user))

    # Add ENTRYPOINT (one layer)
    if entrypoint:
        # TODO(itamar): Consider adding tini as entrypoint also if given
        # Docker CMD without a Docker ENTRYPOINT?
        if full_path_cmd:
            entrypoint[0] = join('/usr/src/app', entrypoint[0])
        if build_context.conf.with_tini_entrypoint:
            entrypoint = ['tini', '--'] + entrypoint
        dockerfile.append(
            'ENTRYPOINT [{}]\n'.format(
                ', '.join(format_docker_cmd(entrypoint))))

    # Add CMD (one layer)
    if cmd:
        if full_path_cmd:
            cmd[0] = join('/usr/src/app', cmd[0])
        dockerfile.append(
            'CMD [{}]\n'.format(', '.join(format_docker_cmd(cmd))))

    # TODO(itamar): write only if changed?
    with open(dockerfile_path, 'w') as dockerfile_f:
        dockerfile_f.writelines(dockerfile)
    docker_build_cmd = [
        'docker', 'build', '-t', docker_image, workspace_dir]
    logger.info('Building docker image "{}" using command {}',
                docker_image, docker_build_cmd)
    run(docker_build_cmd, check=True)
    # TODO(itamar): race condition here
    image_id = get_cached_image_id(docker_image)
    metadata = {
        'image_id': image_id,
        'images': [{
            'name': docker_image,
            'pushed': False,
        }],
    }
    if image_caching_behavior.get('push_image_after_build', False):
        remote_image = get_remote_image_name(name, tag, image_caching_behavior)
        tag_docker_image(image_id, remote_image)
        push_docker_image(remote_image, build_context.conf.docker_push_cmd)
        metadata['images'].append({
            'name': remote_image,
            'pushed': True,
        })
    # Generate ybt_bin scripts
    if ybt_bin_path:
        # Make sure ybt_bin's are created only under bin_path
        assert (build_context.conf.get_bin_path() ==
                commonpath([build_context.conf.get_bin_path(), ybt_bin_path]))

        def format_docker_run_params(params: dict):
            param_strings = []
            if 'container_name' in params:
                param_strings.extend(['--name', params['container_name']])
            if params.get('interactive'):
                param_strings.append('-i')
            if params.get('term'):
                param_strings.append('-t')
            if params.get('rm'):
                param_strings.append('--rm')
            if params.get('daemonize'):
                param_strings.append('-d')
            if params.get('impersonate') and platform.system() == 'Linux':
                param_strings.extend([
                    '-u', '$( id -u ):$( id -g )',
                    '-v', '/etc/passwd:/etc/passwd:ro',
                    '-v', '/etc/group:/etc/group:ro'])
            for port in params['ports']:
                param_strings.extend(['-p', port])
            for volume in params['volumes']:
                param_strings.extend(['-v', volume])
            if params.get('work_dir'):
                param_strings.extend(['-w', params['work_dir']])
            for var, value in params['env'].items():
                param_strings.extend(['-e', '{}="{}"'.format(var, value)])
            return ' '.join(param_strings)

        with open(join(dirname(abspath(__file__)),
                  'ybtbin.sh.tmpl'), 'r') as tmpl_f:
            ybt_bin = tmpl_f.read().format(
                image_name=docker_image, image_id=image_id,
                docker_opts=format_docker_run_params(runtime_params),
                default_opts='$IT' if runtime_params.get('auto_it') else '')
        with open(ybt_bin_path, 'w') as ybt_bin_f:
            ybt_bin_f.write(ybt_bin)
        os.chmod(ybt_bin_path, 0o755)
        metadata['ybt_bin'] = ybt_bin_path
    return metadata


def base_image_caching_behavior(conf: Config, **kwargs):
    base_dict = {
        'skip_build_if_cached': not conf.build_base_images,
        'pull_if_not_cached': not conf.offline,
        'pull_if_cached': conf.force_pull and not conf.offline,
        'fail_build_if_pull_failed': not conf.offline,
        'allow_build_if_not_cached': conf.build_base_images,
        'push_image_after_build': conf.push and not conf.offline,
    }
    base_dict.update(kwargs)
    return base_dict


def deployable_caching_behavior(conf: Config, **kwargs):
    base_dict = {
        'push_image_after_build': conf.push and not conf.offline,
    }
    base_dict.update(kwargs)
    return base_dict



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Docker tests
~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import pytest

from subprocess import PIPE

from .buildcontext import BuildContext
from .graph import populate_targets_graph, topological_sort
from .yabt import cmd_build


slow = pytest.mark.skipif(not pytest.config.getoption('--with-slow'),
                          reason='need --with-slow option to run')


@slow
@pytest.mark.usefixtures('in_simple_project')
def test_run_in_buildenv(basic_conf):
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    for target_name in topological_sort(build_context.target_graph):
        target = build_context.targets[target_name]
        build_context.build_target(target)
    result = build_context.run_in_buildenv(
        'app:flask-hello', ['pip', 'freeze'], stdout=PIPE, stderr=PIPE)
    assert 0 == result.returncode
    for package in [
            b'Flask',
            b'itsdangerous',
            b'Jinja2',
            b'MarkupSafe',
            b'Werkzeug',
            ]:
        assert package in result.stdout


@slow
@pytest.mark.usefixtures('in_simple_project')
def test_ybt_bin_generation(basic_conf):
    basic_conf.targets = ['app:flask-app']
    cmd_build(basic_conf)
    with open('ybt_bin/app/flask-app', 'r') as app_ybt_bin:
        assert ('docker run --name my-flask-app -p 5555:5000'
                in app_ybt_bin.read())


@slow
@pytest.mark.usefixtures('in_pkgmgrs_project')
def test_package_managers_install_order(basic_conf):
    basic_conf.targets = [':the-image']
    cmd_build(basic_conf)
    exp_dockerfile = [
        'FROM python:3\n',
        'ARG DEBIAN_FRONTEND=noninteractive\n',
        'USER root\n',
        'ENV FOO="BAR" PATH="${PATH}:/foo/bar:/ham:/spam" TEST="1"\n',
        'RUN apt-key adv --keyserver hkp://keyserver.ubuntu.com:80 --recv '
        '80F70E11F0F0D5F10CB20E62F5DA5F09C3173AA6\n',
        'COPY the-image.list /etc/apt/sources.list.d/\n',
        'RUN apt-get update -y && apt-get install --no-install-recommends -y '
        'apt-transport-https curl wget && rm -rf /var/lib/apt/lists/*\n',
        'COPY packages1 /tmp/install1\n',
        'RUN tar -xf /tmp/install1/node.tar.gz -C /tmp/install1 && '
        'cd /tmp/install1/node && ./install-nodejs.sh && cd / && '
        'rm -rf /tmp/install1\n',
        'RUN apt-get update -y && apt-get install --no-install-recommends -y '
        'ruby2.2 ruby2.2-dev && rm -rf /var/lib/apt/lists/*\n',
        'COPY requirements_pip_1.txt /usr/src/\n',
        'RUN pip install --no-cache-dir --upgrade pip && '
        'pip install --no-cache-dir -r /usr/src/requirements_pip_1.txt\n',
        'RUN npm install left-pad --global\n',
        'RUN gem install compass\n',
        'COPY requirements_pip_2.txt /usr/src/\n',
        'RUN pip install --no-cache-dir -r /usr/src/requirements_pip_2.txt\n',
        'WORKDIR /usr/src/app\n',
        'USER root\n',
        'CMD ["foo"]\n',
    ]
    with open('yabtwork/DockerBuilder/the-image_latest/Dockerfile',
              'r') as dockerfile:
        assert exp_dockerfile == dockerfile.readlines()


@slow
@pytest.mark.usefixtures('in_pkgmgrs_project')
def test_generate_needed_lists(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = [':another-image']
    populate_targets_graph(build_context, basic_conf)
    for target_name in topological_sort(build_context.target_graph):
        target = build_context.targets[target_name]
        build_context.build_target(target)
    result = build_context.run_in_buildenv(
        ':another-image', ['ls', '/etc/apt/sources.list.d/'],
        stdout=PIPE, stderr=PIPE)
    assert 0 == result.returncode
    for file in [
            b'another-image.list',
            b'the-image.list',
            ]:
        assert file in result.stdout



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Build context module
~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from collections import defaultdict
import json
import os
from os.path import join
import platform
import threading

from ostrich.utils.proc import run
from ostrich.utils.text import get_safe_path

from .config import Config
from .docker import format_qualified_image_name
from .extend import Plugin
from .graph import get_descendants, topological_sort
from .logging import make_logger
from .target_extraction import extractor
from .target_utils import split_build_module, Target


logger = make_logger(__name__)


class BuildContext:
    """Build Context class.

    The framework is designed to operate with a single instance of the
    BuildContext alive, but this is not enforced using a singleton pattern,
    because there's no particular reason to do so...

    While not yet implemented, it is designed to supported concurrent parsing
    of multiple build files using the same build context.
    The design principle to enable this:
    - The "driver" (e.g. main thread in main process) is responsible for
      dispatching build file parsers with copies or references of the build
      context (e.g. on other threads / processes / machines?).
    - Parsers, with their own build context, populate partial target maps.
    - The driver collects the partial results, and is responsible to merge
      them consistently and safely.
    - This will worth it, of course, only if serializing the state to a parser
      and serializing the partial result and merging it back is MUCH cheaper
      than actually doing the parsing, because if it's not, it would be better
      to do it all in the driver (much simpler...) - which is the reason this
      is not yet implemented.
    """

    def __init__(self, conf: Config):
        self.conf = conf
        # A *thread-safe* targets map
        self.targets = {}
        # A *thread-safe* map from build module to set of target names
        # that were extracted from that build module
        self.targets_by_module = defaultdict(set)
        # A *thread-safe* set of processed build-files
        self.processed_build_files = set()
        # Target graph is *not necessarily thread-safe*!
        self.target_graph = None
        # # A *thread-safe* map from BuildEnv name to qualified Docker image
        # #  name for that BuildEnv
        # self.buildenv_images = {}
        # A dictionary for collecting metadata on build artifacts
        self.artifacts_metadata = {}
        self.context_lock = threading.Lock()

    def get_workspace(self, *parts) -> str:
        """Return a path to a private workspace dir.
           Create sub-tree of dirs using strings from `parts` inside workspace,
           and return full path to innermost directory.

        Upon returning successfully, the directory will exist (potentially
        changed to a safe FS name), even if it didn't exist before, including
        any intermediate parent directories.
        """
        workspace_dir = os.path.join(self.conf.get_workspace_path(),
                                     *(get_safe_path(part)
                                       for part in parts))
        if not os.path.isdir(workspace_dir):
            # exist_ok=True in case of concurrent creation of the same
            # workspace
            os.makedirs(workspace_dir, exist_ok=True)
        return workspace_dir

    def get_bin_dir(self, build_module: str) -> str:
        """Return a path to the binaries dir for a build module dir.
           Create sub-tree of missing dirs as needed, and return full path
           to innermost directory.
        """
        bin_dir = os.path.join(self.conf.get_bin_path(), build_module)
        if not os.path.isdir(bin_dir):
            # exist_ok=True in case of concurrent creation of the same bin dir
            os.makedirs(bin_dir, exist_ok=True)
        return bin_dir

    def walk_target_graph(self, target_names: iter):
        """Generate entire target sub-tree for given `target_names` in order.

        Yields target *instances* in the sub-tree, including the nodes given.
        """
        for target_name in target_names:
            yield self.targets[target_name]
            yield from self.walk_target_graph(
                self.target_graph.neighbors_iter(target_name))

    def walk_target_deps_topological_order(self, target: Target):
        all_deps = get_descendants(self.target_graph, target.name)
        for dep_name in topological_sort(self.target_graph):
            if dep_name in all_deps:
                yield self.targets[dep_name]

    def register_target(self, target: Target):
        """Register a `target` instance in this build context.

        A registered target is saved in the `targets` map and in the
        `targets_by_module` map, but is not added to the target graph until
        target extraction is completed (thread safety considerations).
        """
        if target.name in self.targets:
            first = self.targets[target.name]
            raise NameError(
                'Target with name "{0.name}" ({0.builder_name} from module '
                '"{1}") already exists - defined first as '
                '{2.builder_name} in module "{3}"'.format(
                    target, split_build_module(target.name),
                    first, split_build_module(first.name)))
        self.targets[target.name] = target
        self.targets_by_module[split_build_module(target.name)].add(
            target.name)

    def remove_target(self, target_name: str):
        """Remove (unregister) a `target` from this build context.

        Removes the target instance with the given name, if it exists,
        from both the `targets` map and the `targets_by_module` map.

        Doesn't do anything if no target with that name is found.

        Doesn't touch the target graph, if it exists.
        """
        if target_name in self.targets:
            del self.targets[target_name]
        if split_build_module(target_name) in self.targets_by_module:
            self.targets_by_module[split_build_module(target_name)].remove(
                target_name)

    def get_target_extraction_context(self, build_file_path: str) -> dict:
        """Return a build file parser target extraction context.

        The target extraction context is a build-file-specific mapping from
        builder-name to target extraction function,
        for every registered builder.
        """
        extraction_context = {}
        for name, builder in Plugin.builders.items():
            extraction_context[name] = extractor(name, builder,
                                                 build_file_path, self)
        return extraction_context

    # def register_buildenv_image(self, name: str, docker_image: str):
    #     """Register a named BuildEnv Docker image in this build context."""
    #     self.buildenv_images[name] = docker_image

    def run_in_buildenv(
            self, buildenv_target_name: str, cmd: list, cmd_env: dict=None,
            work_dir: str=None, auto_uid: bool=True, **kwargs):
        """Run a command in a named BuildEnv Docker image.

        :param buildenv_target_name: A named Docker image target in which the
                                     command should be run.
        :param cmd: The command to run, as you'd pass to subprocess.run()
        :param cmd_env: A dictionary of environment variables for the command.
        :param work_dir: A different work dir to run in.
                         Either absolute path, or relative to project root.
        :param auto_uid: Whether to run as the active uid:gid, or as root.
        :param kwargs: Extra keyword arguments that are passed to the
                        subprocess.run() call that runs the BuildEnv container
                        (for, e.g. timeout arg, stdout/err redirection, etc.)

        :raises KeyError: If named BuildEnv is not a registered BuildEnv image
        """
        buildenv_target = self.targets[buildenv_target_name]
        # TODO(itamar): Assert that buildenv_target is up to date
        redirection = any(
            stream_key in kwargs
            for stream_key in ('stdin', 'stdout', 'stderr', 'input'))
        docker_run = ['docker', 'run']
        if not self.conf.non_interactive:
            docker_run.append('-i')
        if not redirection:
            docker_run.append('-t')
        project_vol = (self.conf.docker_volume if self.conf.docker_volume else
                       self.conf.project_root)
        docker_run.extend([
            '--rm',
            '-v', project_vol + ':/project',
            '-w', join('/project', work_dir) if work_dir else '/project',
        ])
        if cmd_env:
            for key, value in cmd_env.items():
                # TODO(itamar): escaping
                docker_run.extend(['-e', '{}={}'.format(key, value)])
        if platform.system() == 'Linux' and auto_uid:
            # Fix permissions for bind-mounted project dir
            # The fix is not needed when using Docker machine on OS X with
            # VirtualBox or xhyve, because it is somehow taken care of by
            # the shared folder thingie...
            docker_run.extend(['-u', '{}:{}'.format(os.getuid(), os.getgid())])
        docker_run.append(format_qualified_image_name(buildenv_target))
        docker_run.extend(cmd)
        logger.info('Running command in build env "{}" using command {}',
                    buildenv_target_name, docker_run)
        return run(docker_run, check=True, **kwargs)

    def build_target(self, target: Target):
        """Invoke the builder function for a target."""
        builder = Plugin.builders[target.builder_name]
        if builder.func:
            logger.info('About to invoke the {} builder function for {}',
                        target.builder_name, target)
            builder.func(self, target)
        else:
            logger.warning('Skipping {} builder function for target {} (no '
                           'function registered)', target.builder_name, target)

    def register_target_artifact_metadata(self, target: str, metadata: dict):
        """Register the artifact metadata dictionary for a built target."""
        with self.context_lock:
            target.props.docker_image_id = metadata['image_id']
            self.artifacts_metadata[target.name] = metadata

    def write_artifacts_metadata(self):
        """Write out a JSON file with all built targets artifact metadata,
           if such output file is specified."""
        if self.conf.artifacts_metadata_file:
            logger.info('Writing artifacts metadata to file "%s"',
                        self.conf.artifacts_metadata_file)
            with open(self.conf.artifacts_metadata_file, 'w') as fp:
                json.dump(self.artifacts_metadata, fp)



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Grunt builder
~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import os

from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..logging import make_logger
from ..target_utils import split_build_module
from ..utils import yprint


logger = make_logger(__name__)


register_builder_sig(
    'Grunt',
    [('in_buildenv', PT.Target, None),
     ('grunt_tasks', PT.StrList, None),
     ('cmd_env', None),
     ('work_dir', None),
     ])


@register_build_func('Grunt')
def grunt_builder(build_context, target):
    yprint(build_context.conf, 'Build (run) Grunt', target)
    # npm_links = [dep.props.package for dep in
    #              build_context.walk_target_deps_topological_order(target)
    #              if ('npm-installable' in dep.tags and
    #                  dep.props.global_install)]
    work_dir = (target.props.work_dir if target.props.work_dir
                else split_build_module(target.name))
    logger.debug('Cleaning ".tmp" directory')
    build_context.run_in_buildenv(
        target.props.in_buildenv, ['rm', '-rf', '.tmp'],
        target.props.cmd_env, work_dir, auto_uid=False)
    yprint(build_context.conf, 'Running "npm install" in {}'.format(work_dir))
    build_context.run_in_buildenv(
        target.props.in_buildenv, ['npm', 'install'],
        target.props.cmd_env, work_dir, auto_uid=False)
    yprint(build_context.conf,
           'Running "bower install" in {}'.format(work_dir))
    build_context.run_in_buildenv(
        target.props.in_buildenv, ['bower', 'install', '--allow-root'],
        target.props.cmd_env, work_dir, auto_uid=False)
    grunt_cmd = ['grunt'] + target.props.grunt_tasks
    yprint(build_context.conf,
           'Running "{}" in {}'.format(' '.join(grunt_cmd), work_dir))
    build_context.run_in_buildenv(
        target.props.in_buildenv, grunt_cmd,
        target.props.cmd_env, work_dir, auto_uid=False)
    logger.debug('Fix grunt output ownership')
    # TODO(itamar): Don't hardcode output dir here... ("dist")
    build_context.run_in_buildenv(
        target.props.in_buildenv,
        ['chown', '-R', '{}:{}'.format(os.getuid(), os.getgid()), 'dist'],
        target.props.cmd_env, work_dir, auto_uid=False)
    # TODO(itamar): way to describe the artifacts of the external command,
    # so it can be used by dependent targets, and cached in some smart way


@register_manipulate_target_hook('Grunt')
def grunt_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt target graph
~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import networkx
from networkx.algorithms import dag

from .buildfile_parser import process_build_file
from .config import Config
from .logging import make_logger
from .target_utils import generate_build_modules
from .target_utils import parse_target_selectors
from .utils import yprint


logger = make_logger(__name__)


def stable_reverse_topological_sort(graph):
    """Return a list of nodes in topological sort order.

    This topological sort is a **unique** permutation of the nodes
    such that an edge from u to v implies that u appears before v in the
    topological sort order.

    Parameters
    ----------
    graph : NetworkX digraph
            A directed graph

    Raises
    ------
    NetworkXError
        Topological sort is defined for directed graphs only. If the
        graph G is undirected, a NetworkXError is raised.
    NetworkXUnfeasible
        If G is not a directed acyclic graph (DAG) no topological sort
        exists and a NetworkXUnfeasible exception is raised.

    Notes
    -----
    - This algorithm is based on a description and proof in
      The Algorithm Design Manual [1]_ .
    - This implementation is modified from networkx 1.11 implementation [2]_
      to achieve stability, support only reverse (allows yielding instead of
      returning a list), and remove the `nbunch` argument (had no use for it).

    See also
    --------
    is_directed_acyclic_graph

    References
    ----------
    .. [1] Skiena, S. S. The Algorithm Design Manual  (Springer-Verlag, 1998).
        http://www.amazon.com/exec/obidos/ASIN/0387948600/ref=ase_thealgorithmrepo/
    .. [2] networkx on GitHub
        https://github.com/networkx/networkx/blob/8358afac209c00b7feb3e81c901098852a9413b3/networkx/algorithms/dag.py#L88-L168
    """
    if not graph.is_directed():
        raise networkx.NetworkXError(
            'Topological sort not defined on undirected graphs.')

    # nonrecursive version
    seen = set()
    explored = set()

    for v in sorted(graph.nodes_iter()):
        if v in explored:
            continue
        fringe = [v]  # nodes yet to look at
        while fringe:
            w = fringe[-1]  # depth first search
            if w in explored:  # already looked down this branch
                fringe.pop()
                continue
            seen.add(w)     # mark as seen
            # Check successors for cycles and for new nodes
            new_nodes = []
            for n in sorted(graph[w]):
                if n not in explored:
                    if n in seen:  # CYCLE!! OH NOOOO!!
                        raise networkx.NetworkXUnfeasible(
                            'Graph contains a cycle.')
                    new_nodes.append(n)
            if new_nodes:   # Add new_nodes to fringe
                fringe.extend(new_nodes)
            else:           # No new nodes so w is fully explored
                explored.add(w)
                yield w
                fringe.pop()  # done considering this node


def build_target_dep_graph(build_context, unused_conf: Config):
    build_context.target_graph = networkx.DiGraph()
    for target_name, target in build_context.targets.items():
        build_context.target_graph.add_node(target_name)
        for dep in target.deps:
            build_context.target_graph.add_edge(target_name, dep)


def populate_targets_graph(build_context, conf: Config):
    # Process project root build file
    process_build_file(conf.get_project_build_file(), build_context, conf)
    targets_to_prune = set(build_context.targets.keys())
    if conf.targets:
        yprint(build_context.conf, 'targets:', conf.targets)
        # TODO(itamar): Figure out how to support a target selector that is a
        #   parent directory which isn't a build module, but contains build
        #   modules (e.g., `ybt tree yapi` from the `dag` test root).
        seeds = parse_target_selectors(conf.targets, conf)
        yprint(build_context.conf, 'seeds', seeds)
    else:
        default_target = ':{}'.format(conf.default_target_name)
        logger.info('searching for default target {}', default_target)
        if default_target not in build_context.targets:
            raise RuntimeError(
                'No default target found, and no target selector specified')
        seeds = [default_target]

    # Crawl rest of project from seeds
    seeds_used_for_extending = set()
    for seed in seeds:
        if seed in build_context.targets:
            #
            if seed in targets_to_prune:
                targets_to_prune.remove(seed)
            if seed not in seeds_used_for_extending:
                # Avoid infinite loop in case of cyclic dependencies
                seeds.extend(build_context.targets[seed].deps)
                if build_context.targets[seed].buildenv:
                    seeds.append(build_context.targets[seed].buildenv)
                seeds_used_for_extending.add(seed)
        else:
            if seed == '**:*':
                # Adding all build modules under current working directory as
                # seeds
                seeds.extend(generate_build_modules('.', conf))
                continue
            build_module, target_name = seed.split(':', 1)
            process_build_file(conf.get_build_file_path(build_module),
                               build_context, conf)
            # Parsed build file with this seed target - add its dependencies as
            # seeds
            if target_name == '*':
                # It's a wildcard - add all targets from build module
                # (and skip adding to targets_to_prune altogether)
                for module_target in (
                        build_context.targets_by_module[build_module]):
                    seeds.extend(
                        build_context.targets[module_target].deps)
                    if build_context.targets[module_target].buildenv:
                        seeds.append(
                            build_context.targets[module_target].buildenv)
            else:
                if seed not in build_context.targets:
                    raise RuntimeError(
                        'Don\'t know how to make `{}\''.format(seed))
                #
                for module_target in (
                        build_context.targets_by_module[build_module]):
                    targets_to_prune.add(module_target)
                targets_to_prune.remove(seed)
                seeds.extend(build_context.targets[seed].deps)
                if build_context.targets[seed].buildenv:
                    seeds.append(build_context.targets[seed].buildenv)
        # TODO(itamar): Write tests that pruning is *ALWAYS* correct!
        # e.g., not pruning things it shouldn't (like when targets are in prune
        # list when loaded initially, but should be removed later because a
        # target loaded later required them).

    # Pruning, after parsing is done
    # (first, adding targets that are tagged as "prune-me" to prune list)
    targets_to_prune.update(
        target_name
        for target_name, target in build_context.targets.items()
        if 'prune-me' in target.tags)
    for target_name in targets_to_prune:
        build_context.remove_target(target_name)

    build_target_dep_graph(build_context, conf)
    assert dag.is_directed_acyclic_graph(build_context.target_graph)


def topological_sort(graph: networkx.DiGraph):
    yield from stable_reverse_topological_sort(graph)


def get_descendants(graph: networkx.DiGraph, source):
    """Return all nodes reachable from `source` in `graph`."""
    return dag.descendants(graph, source)



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

:author: Zohar Rimon
"""


import os
from os.path import isdir, isfile, join
import shutil

import pytest

from . import proto
from ..buildcontext import BuildContext
from ..graph import populate_targets_graph, topological_sort
from ..utils import yprint


slow = pytest.mark.skipif(not pytest.config.getoption('--with-slow'),
                          reason='need --with-slow option to run')


def clear_output():
    try:
        shutil.rmtree('build')
    except FileNotFoundError:
        pass


@slow
@pytest.mark.usefixtures('in_prototest_project')
def test_proto_builder(basic_conf):
    clear_output()
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['app:hello-proto']
    populate_targets_graph(build_context, basic_conf)
    for target_name in topological_sort(build_context.target_graph):
        target = build_context.targets[target_name]
        if target.buildenv:
            buildenv = build_context.targets[target.buildenv]
            for dep in build_context.walk_target_deps_topological_order(
                    buildenv):
                build_context.build_target(dep)
            build_context.build_target(buildenv)
        build_context.build_target(target)
    assert isdir('build')
    assert isdir(join('build', 'gen'))
    assert isdir(join('build', 'gen', 'proto'))
    assert isfile(join('build', 'gen', 'proto', '__init__.py'))
    assert isdir(join('build', 'gen', 'proto', 'app'))
    assert isfile(join('build', 'gen', 'proto', 'app', '__init__.py'))
    for exp_gen_fname in [
        'hello.pb.cc',
        'hello.pb.h',
        'hello_pb2.py'
    ]:
        assert isfile(join('build', 'gen', 'proto', 'app', exp_gen_fname))
    clear_output()



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Python Builders
~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from .dockerapp import build_app_docker_and_bin, register_app_builder_sig
from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..logging import make_logger
from ..utils import yprint


logger = make_logger(__name__)


register_builder_sig(
    'PythonPackage',
    [('package', PT.str),
     ('version', PT.str, None),
     ('pip', PT.str, 'pip2'),
     ])


@register_build_func('PythonPackage')
def python_package_builder(build_context, target):
    yprint(build_context.conf, 'Fetch and cache PyPI package', target)


@register_manipulate_target_hook('PythonPackage')
def python_package_manipulate_target(build_context, target):
    target.tags.add('pip-installable')


register_builder_sig(
    'Python',
    [('sources', PT.FileList, None),
     ('data', PT.FileList, None),
     ])


@register_build_func('Python')
def python_app_builder(build_context, target):
    yprint(build_context.conf, 'Build Python', target)
    target.artifacts['app'].extend(target.props.sources)
    target.artifacts['app'].extend(target.props.data)


register_app_builder_sig('PythonApp', [('main', PT.File)])


@register_manipulate_target_hook('PythonApp')
def python_app_manipulate_target(build_context, target):
    logger.debug('Injecting "{}" to deps of {}',
                 target.props.base_image, target)
    target.deps.append(target.props.base_image)


@register_build_func('PythonApp')
def python_app_builder(build_context, target):
    yprint(build_context.conf, 'Build PythonApp', target)
    if target.props.main not in target.artifacts['app']:
        target.artifacts['app'].append(target.props.main)
    build_app_docker_and_bin(
        build_context, target, entrypoint=[target.props.main])



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt DockerApp Builder
~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from os.path import join

from ostrich.utils.collections import listify

from ..docker import build_docker_image, get_image_name
from ..extend import PropType as PT, register_builder_sig
from .. import target_utils


def register_app_builder_sig(builder_name, sig=None, docstring=None):
    register_builder_sig(
        builder_name,
        listify(sig) + [
            ('base_image', PT.Target),
            ('image_name', PT.str, None),
            ('image_tag', PT.str, 'latest'),
            ('full_path_cmd', PT.bool, False),
            ('work_dir', PT.str, '/usr/src/app'),
            ('env', PT.dict, None),
            ('distro', PT.dict, None),
            ('image_caching_behavior', PT.dict, None),
            ('truncate_common_parent', PT.str, None),
            ('build_user', PT.str, None),
            ('run_user', PT.str, None),
            ('docker_labels', PT.dict, None),
        ],
        docstring)


def build_app_docker_and_bin(build_context, target, **kwargs):
    build_module, bin_name = target_utils.split(target.name)
    ybt_bin_path = join(build_context.get_bin_dir(build_module), bin_name)
    metadata = build_docker_image(
        build_context,
        name=get_image_name(target),
        tag=target.props.image_tag,
        base_image=build_context.targets[target.props.base_image],
        deps=build_context.walk_target_deps_topological_order(target),
        env=target.props.env,
        work_dir=target.props.work_dir,
        truncate_common_parent=target.props.truncate_common_parent,
        entrypoint=kwargs.get('entrypoint'),
        full_path_cmd=target.props.full_path_cmd,
        distro=target.props.distro,
        image_caching_behavior=target.props.image_caching_behavior,
        runtime_params=target.props.runtime_params,
        ybt_bin_path=ybt_bin_path,
        build_user=target.props.build_user,
        run_user=target.props.run_user,
        labels=target.props.docker_labels)
    build_context.register_target_artifact_metadata(target, metadata)



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt utils
~~~~~~~~~~

:author: Itamar Ostricher
"""


import os
from os.path import isdir, isfile, join, normpath, relpath, split
import shutil
import sys

from ostrich.utils.path import commonpath

from .compat import scandir
from .logging import make_logger


logger = make_logger(__name__)


def rmtree(path: str):
    """Forcibly remove directory tree.
       Fail silently if base dir doesn't exist."""
    try:
        shutil.rmtree(path)
    except FileNotFoundError:
        pass


def link_func(src: str, dst: str):
    try:
        os.link(src, dst)
    except FileExistsError:
        pass


def link_node(abs_src: str, abs_dest: str):
    """Sync source node (file / dir) to destination path using hard links."""
    if isfile(abs_src):
        # sync file by linking it to dest
        dest_parent_dir = split(abs_dest)[0]
        if not isdir(dest_parent_dir):
            # exist_ok=True in case of concurrent creation of the same
            # parent dir
            os.makedirs(dest_parent_dir, exist_ok=True)
        link_func(abs_src, abs_dest)
    elif isdir(abs_src):
        # sync dir by recursively linking files under it to dest
        shutil.copytree(abs_src, abs_dest,
                        copy_function=link_func,
                        ignore=shutil.ignore_patterns('.git'))
    else:
        raise FileNotFoundError(abs_src)


def link_artifacts(artifacts: set, workspace_src_dir: str,
                   common_parent: str, conf):
    """Sync the list of files and directories in `artifacts` to destination
       directory specified by `workspace_src_dir`.

    "Sync" in the sense that every file given in `artifacts` will be
    hard-linked under `workspace_src_dir` after this function returns, and no
    other files will exist under `workspace_src_dir`.

    For directories in `artifacts`, hard-links of contained files are
    created recursively.

    All paths in `artifacts`, and the `workspace_src_dir`, must be relative
    to `conf.project_root`.

    If `workspace_src_dir` exists before calling this function, it is removed
    before syncing.

    If `common_parent` is given, and it is a common parent directory of all
    `artifacts`, then the `commonm_parent` part is truncated from the
    sync'ed files destination path under `workspace_src_dir`.

    :raises FileNotFoundError: If `artifacts` contains files or directories
                               that do not exist.

    :raises ValueError: If `common_parent` is given (not `None`), but is *NOT*
                        a common parent of all `artifacts`.
    """
    norm_dir = normpath(workspace_src_dir)
    if norm_dir not in conf.deleted_dirs:
        conf.deleted_dirs.add(norm_dir)
        rmtree(norm_dir)
    if common_parent:
        common_parent = normpath(common_parent)
        base_dir = commonpath(list(artifacts) + [common_parent])
        if base_dir != common_parent:
            raise ValueError('{} is not the common parent of all target '
                             'sources and data'.format(common_parent))
        logger.debug(
            'Rebasing files in image relative to common parent dir {}',
            base_dir)
    else:
        base_dir = ''
    num_linked = 0
    for src in artifacts:
        abs_src = join(conf.project_root, src)
        abs_dest = join(conf.project_root, workspace_src_dir,
                        relpath(src, base_dir))
        link_node(abs_src, abs_dest)
        num_linked += 1
    return num_linked


def norm_proj_path(path, build_module):
    """Return a normalized path for the `path` observed in `build_module`.

    The normalized path is "normalized" (in the `os.path.normpath` sense),
    and relative from the project root directory.

    Supports making references from project root directory by prefixing the
    path with "//".

    :raises ValueError: If path references outside the project sandbox.
    """
    if path == '//':
        return ''

    if path.startswith('//'):
        norm = normpath(path[2:])
        if norm[0] in ('.', '/', '\\'):
            raise ValueError("Invalid path: `{}'".format(path))
        return norm

    if path.startswith('/'):
        raise ValueError("Invalid path: `{}' - use '//' to start from "
                         "project root".format(path))

    norm = normpath(join(build_module, path))
    if norm == '.':
        return ''
    if norm.startswith('..'):
        raise ValueError(
            "Invalid path `{}' - must remain inside project sandbox"
            .format(path))
    return norm


def search_for_parent_dir(start_at: str=None, with_files: set=None,
                          with_dirs: set=None) -> str:
    """Return absolute path of first parent directory of `start_at` that
       contains a file named `BUILD_PROJ_FILE` (including `start_at`).

    If `start_at` not specified, start at current working directory.

    :param start_at: Initial path for searching for the project build file.

    Returns `None` upon reaching FS root without finding a project buildfile.
    """
    if not start_at:
        start_at = os.path.abspath(os.curdir)
    if not with_files:
        with_files = set()
    if not with_dirs:
        with_dirs = set()
    exp_hits = len(with_files) + len(with_dirs)
    while start_at:
        num_hits = 0
        for entry in scandir(start_at):
            if ((entry.is_file() and entry.name in with_files) or
                    (entry.is_dir() and entry.name in with_dirs)):
                num_hits += 1
            if num_hits == exp_hits:
                return start_at
        cur_level = start_at
        start_at = os.path.split(cur_level)[0]
        if os.path.realpath(cur_level) == os.path.realpath(start_at):
            # looped on root once
            break


def yprint(config, *objects, **kwargs):
    if config.verbose:
        print(*objects, **kwargs)
    else:
        logger.info('{}', ' '.join(str(obj) for obj in objects))



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Build File parser
~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import os
import sys
import traceback

import colorama

from .config import Config
from .glob import glob
from .logging import make_logger
from .scm import SourceControl


logger = make_logger(__name__)


def err(msg, *args, **kwargs):
    print(msg.format(*args, **kwargs), file=sys.stderr)


def report_buildfile_error(buildfile_path, unused_conf: Config):
    exc_type, exc, tb = sys.exc_info()
    stack = traceback.extract_tb(tb)
    if len(stack) > 1:
        err('Traceback (most recent call first):')
        for fname, line, unused_func, text in reversed(stack):
            err('  File "{}", line {}',
                buildfile_path if fname == '<string>' else fname, line)
            if text:
                err('    {}', text)
            if fname == '<string>':
                break
    print(colorama.Fore.RED, file=sys.stderr, end='')
    if exc_type == SyntaxError:
        err('Syntax Error in build file {}, line {}:\n  {}\n  {: >{}}',
            buildfile_path, exc.lineno,
            # TODO(itamar): Fix issue with empty exc.text
            # (occurs when keyword arg repeats in call to build func)
            exc.text.strip('\n') if exc.text else '', '^', exc.offset)
    else:
        err('YABT Error: {}', exc)
    print(colorama.Style.RESET_ALL, file=sys.stderr, end='')
    sys.exit(1)


def process_build_file(buildfile_path: str, build_context, conf: Config):
    # TODO(itamar): Write tests that verify that this is really not needed in
    # any scenario... (caused issue when referring to target in yroot using
    # `.:foo` syntax from yroot, but normalizing the target name seemed to
    # resolve this). (test also `@` variants etc.)

    # abs_path = os.path.abspath(buildfile_path)
    if buildfile_path in build_context.processed_build_files:
        logger.debug('Skipping processed build file {}', buildfile_path)
        return
    build_context.processed_build_files.add(buildfile_path)
    logger.info('Processing build file {}', buildfile_path)

    with open(buildfile_path, 'r') as buildfile:
        global_context = globals()
        global_context.update({
            'conf': conf,
            'SCM': conf.scm,
            'Glob': glob,
        })
        curdir = os.getcwd()
        try:
            os.chdir(os.path.dirname(buildfile_path))
            # pylint: disable=exec-used
            exec(buildfile.read(), global_context,
                 build_context.get_target_extraction_context(buildfile_path))
        except:
            report_buildfile_error(buildfile_path, conf)
        finally:
            os.chdir(curdir)



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt C++ Builder
~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from .dockerapp import build_app_docker_and_bin, register_app_builder_sig
from ..extend import (
    PropType as PT, register_build_func, register_manipulate_target_hook)
from ..logging import make_logger
from ..utils import yprint


logger = make_logger(__name__)


register_app_builder_sig('CppApp', [('executable', PT.File)])


@register_manipulate_target_hook('CppApp')
def cpp_app_manipulate_target(build_context, target):
    logger.debug('Injecting "{}" to deps of {}',
                 target.props.base_image, target)
    target.deps.append(target.props.base_image)


@register_build_func('CppApp')
def cpp_bin_builder(build_context, target):
    """Pack a C++ binary as a Docker image with its runtime dependencies.

    TODO(itamar): Dynamically analyze the binary and copy shared objects
    from its buildenv image to the runtime image, unless they're installed.
    """
    yprint(build_context.conf, 'Build CppApp', target)
    if target.props.executable not in target.artifacts['app']:
        target.artifacts['app'].append(target.props.executable)
    build_app_docker_and_bin(
        build_context, target, entrypoint=[target.props.executable])



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Docker Builder
~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..logging import make_logger
from ..docker import build_docker_image, get_image_name
from ..utils import yprint


logger = make_logger(__name__)


# `deps` on ExtDockerImage is a way to tell YABT that this external image
# comes preinstalled with a bunch of deps, so when using a specific external
# image as a base image for a new Docker image target that has common deps
# with those of the extenral image, there's no need to re-add those deps
# during the build of the new image.
# Beyond that, it doesn't mean anything to have "deps" on an external image.
register_builder_sig(
    'ExtDockerImage',
    [('image', PT.str),
     ('tag', PT.str, None),
     ('distro', PT.dict, None),
     ])


@register_build_func('ExtDockerImage')
def ext_docker_image_builder(build_context, target):
    yprint(build_context.conf,
           'Fetch and cache Docker image from registry', target)


register_builder_sig(
    'DockerImage',
    [('start_from', PT.Target),
     ('docker_entrypoint', PT.StrList, None),
     ('docker_cmd', PT.StrList, None),
     ('full_path_cmd', PT.bool, False),
     ('image_name', PT.str, None),
     ('image_tag', PT.str, 'latest'),
     ('work_dir', PT.str, '/usr/src/app'),
     ('env', PT.dict, None),
     ('distro', PT.dict, None),
     ('image_caching_behavior', PT.dict, None),
     ('truncate_common_parent', PT.str, None),
     ('ybt_bin_path', PT.str, None),
     ('build_user', PT.str, None),
     ('run_user', PT.str, None),
     ('docker_labels', PT.dict, None),
     ])


@register_manipulate_target_hook('DockerImage')
def docker_image_manipulate_target(build_context, target):
    logger.debug('Injecting "{}" to deps of {}',
                 target.props.start_from, target)
    target.deps.append(target.props.start_from)


@register_build_func('DockerImage')
def docker_image_builder(build_context, target):
    metadata = build_docker_image(
        build_context,
        name=get_image_name(target),
        tag=target.props.image_tag,
        base_image=build_context.targets[target.props.start_from],
        deps=build_context.walk_target_deps_topological_order(target),
        env=target.props.env,
        work_dir=target.props.work_dir,
        truncate_common_parent=target.props.truncate_common_parent,
        entrypoint=target.props.docker_entrypoint,
        cmd=target.props.docker_cmd,
        full_path_cmd=target.props.full_path_cmd,
        distro=target.props.distro,
        image_caching_behavior=target.props.image_caching_behavior,
        runtime_params=target.props.runtime_params,
        ybt_bin_path=target.props.ybt_bin_path,
        build_user=target.props.build_user,
        run_user=target.props.run_user,
        labels=target.props.docker_labels)
    build_context.register_target_artifact_metadata(target, metadata)



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt config
~~~~~~~~~~~

:author: Itamar Ostricher
"""


import os

from .extend import Plugin
from .logging import configure_logging
from .scm import ScmManager
from .utils import search_for_parent_dir


BUILD_PROJ_FILE = 'YRoot'
YCONFIG_FILE = 'YConfig'
YSETTINGS_FILE = 'YSettings'


class Config:
    """Runtime Config info class"""

    attrs_from_args = frozenset((
        'bin_output_dir', 'build_file_name', 'default_target_name', 'cmd',
        'targets', 'build_base_images', 'builders_workspace_dir', 'force_pull',
        'offline', 'non_interactive', 'push', 'docker_volume', 'verbose',
        'with_tini_entrypoint', 'artifacts_metadata_file',
        'loglevel', 'logtostderr', 'logtostdout',
    ))

    def __init__(self, args, project_root_dir: str, work_dir: str):
        """
        :param project_root_dir: Absolute path to build project root directory.
        :param work_dir: Absolute path to working directory within project.
        """
        for slot in self.attrs_from_args:
            setattr(self, slot, getattr(args, slot))
        self.docker_pull_cmd = args.docker_pull_cmd.split()
        self.docker_push_cmd = args.docker_push_cmd.split()
        configure_logging(self)
        self.project_root = project_root_dir
        self.work_dir = work_dir
        self.scm_provider = str(args.scm_provider).lower()
        self.scm = ScmManager.get_provider(self.scm_provider, self)
        Plugin.load_plugins(self)
        self.deleted_dirs = set()

    def in_yabt_project(self) -> bool:
        return self.project_root is not None

    def get_rel_work_dir(self) -> str:
        return os.path.relpath(self.work_dir, self.project_root)

    def get_project_build_file(self) -> str:
        return os.path.join(self.project_root, BUILD_PROJ_FILE)

    def get_build_file_path(self, build_module) -> str:
        is_root_module = os.path.abspath(build_module) == self.project_root
        return os.path.join(
            self.project_root, build_module,
            BUILD_PROJ_FILE if is_root_module else self.build_file_name)

    def get_workspace_path(self) -> str:
        return os.path.join(self.project_root, self.builders_workspace_dir)

    def get_bin_path(self) -> str:
        return os.path.join(self.project_root, self.bin_output_dir)

    def host_to_buildenv_path(self, host_path: str) -> str:
        return os.path.join(
            '/project',
            os.path.relpath(os.path.abspath(host_path), self.project_root))



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt source control subsystem
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import os
import sys

import git

from ..config import Config
from ..logging import make_logger
from ..scm import register_scm_provider
from ..utils import search_for_parent_dir


logger = make_logger(__name__)


@register_scm_provider('git')
class GitSCM:
    """Git SCM implementation."""

    def __init__(self, conf: Config):
        self.repo_dir = search_for_parent_dir(conf.project_root,
                                              with_dirs=set(['.git']))
        if not self.repo_dir:
            print('fatal: "git" scm-provider option given, but not a git '
                  'repository (or any of the parent directories): .git')
            sys.exit(1)
        self._repo = git.Repo(self.repo_dir)
        logger.debug('Initialized Git Repo object at {}', self.repo_dir)
        self.revision = None

    def get_revision(self):
        """Return Git hash of HEAD commit of active repository."""
        if self.revision:
            return self.revision

        if 'GIT_COMMIT' in os.environ:
            self.revision = os.environ['GIT_COMMIT']
        else:
            self.revision = self._repo.head.commit.hexsha
        return self.revision



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt extension subsystem tests
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import pytest

from .extend import (
    Plugin, PropType as PT, register_builder_sig, register_build_func)


def test_load_plugins(basic_conf):
    # getting the `basic_conf` fixture makes sure that plugins are loaded
    assert len(Plugin.builders) > 0
    assert 'Alias' in Plugin.builders


def test_non_default_args_after_default_args_error():
    Plugin.remove_builder('Spam')
    with pytest.raises(SyntaxError) as excinfo:
        register_builder_sig('Spam', ['foo',
                                      ('cats', PT.TargetList, None),
                                      ('bar'),
                                      ])
    assert (
        'non-default argument follows default argument' in str(excinfo.value))


def test_dup_builder_sig_error():
    Plugin.remove_builder('Spam')
    register_builder_sig('Spam', [])
    with pytest.raises(KeyError) as excinfo:
        register_builder_sig('Spam', [])
    assert 'Spam already registered a signature!' in str(excinfo.value)


def test_dup_builder_func_error():
    Plugin.remove_builder('Spam')

    def spam_builder_1(build_context, target):
        pass

    register_build_func('Spam')(spam_builder_1)
    with pytest.raises(KeyError) as excinfo:
        register_build_func('Spam')(spam_builder_1)
    assert 'Spam already registered a build function!' in str(excinfo.value)


def test_register_build_sig_duplicate_arg():
    Plugin.remove_builder('Spam')
    with pytest.raises(SyntaxError) as excinfo:
        register_builder_sig('Spam', [('deps', PT.TargetList, None)])
    assert ("duplicate argument 'deps' in function definition"
            in str(excinfo.value))
    Plugin.remove_builder('Spam')
    with pytest.raises(SyntaxError) as excinfo:
        register_builder_sig('Spam', [('name', PT.TargetName)])
    assert ("duplicate argument 'name' in function definition"
            in str(excinfo.value))


def test_builder_sig_untyped_default_value():
    Plugin.remove_builder('Spam')
    register_builder_sig('Spam', [('foo', 'default foo')])
    assert PT.untyped == Plugin.builders['Spam'].sig['foo'].type
    assert 'default foo' == Plugin.builders['Spam'].sig['foo'].default



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Target utils tests
~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import pytest

from .target_utils import norm_name


def test_norm_name_abs_ref():
    """Test norm_name abs ref to targets in nested modules."""
    assert 'cat/foo:bar' == norm_name('cat', '//cat/foo:bar')
    assert 'cat:bar' == norm_name('cat', '//cat:bar')
    assert 'dog:bark' == norm_name('cat', '//dog:bark')
    assert 'stuff:*' == norm_name('cat', '//stuff:*')
    assert 'stuff/**:*' == norm_name('cat', '//stuff/**:*')


def test_norm_name_abs_ref_top_level():
    """Test norm_name abs ref to targets in top level module."""
    assert ':spam' == norm_name('cat', '//:spam')
    assert ':*' == norm_name('cat', '//:*')
    assert '**:*' == norm_name('cat', '//**:*')


def test_norm_name_rel_ref_sub():
    """Test norm_name rel ref to sub-module."""
    assert 'cat/foo:bar' == norm_name('cat', 'foo:bar')
    assert 'cat/foo:bar' == norm_name('cat', './foo:bar')
    assert 'cat/foo:*' == norm_name('cat', 'foo:*')
    assert 'cat/foo/**:*' == norm_name('cat', 'foo/**:*')


def test_norm_name_rel_ref_local():
    """Test norm_name rel ref to same module."""
    assert 'cat:bar' == norm_name('cat', ':bar')
    assert 'cat:bar' == norm_name('cat', '.:bar')
    assert 'cat:bar' == norm_name('cat', './:bar')
    assert 'cat:*' == norm_name('cat', ':*')
    assert 'cat/**:*' == norm_name('cat', '**:*')


def test_norm_name_rel_ref_sub():
    """Test norm_name rel ref to a sibling module."""
    assert 'dog:bark' == norm_name('cat', '../dog:bark')
    assert 'dog:bark' == norm_name('cat', '../dog/.:bark')
    assert 'dog:bark' == norm_name('cat', './../dog:bark')
    assert 'dog:bark' == norm_name('cat', '.././dog:bark')
    assert 'dog:*' == norm_name('cat', '../dog:*')
    assert 'dog/**:*' == norm_name('cat', '../dog/**:*')


def test_norm_name_escape_sandbox_error():
    """Test norm_name various sandbox-escaping scenarios."""
    for bad_name in [
            '///cat/foo:bar',
            '//../cat/foo:bar',
            '/etc/passwd:meow',
            '../..:paw']:
        with pytest.raises(ValueError) as excinfo:
            norm_name('cat', bad_name)
    assert 'must remain inside project sandbox' in str(excinfo.value)


def test_norm_name_unqualified_error():
    """Test norm_name when given unqualified target name."""
    with pytest.raises(ValueError) as excinfo:
        norm_name('cat', 'poops')
    assert ('Must provide fully-qualified target name (with `:\') to avoid '
            'possible ambiguity' in str(excinfo.value))



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt cli module
~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from importlib.machinery import SourceFileLoader
import os

import argcomplete
import colorama
import configargparse

from .config import BUILD_PROJ_FILE, Config, YCONFIG_FILE, YSETTINGS_FILE
from .utils import search_for_parent_dir


PARSER = None
KNOWN_LOG_LEVELS = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
LOG_LEVELS_CHOICES = KNOWN_LOG_LEVELS + [level.lower()
                                         for level in KNOWN_LOG_LEVELS]


def make_parser(project_config_file: str) -> configargparse.ArgumentParser:
    """Return the argument parser.

    :param project_config_file: Absolute path to project-specific config file.

    If cached parser already exists - return it immediately.
    Otherwise, initialize a new `ConfigArgParser` that is able to take default
    values from a hierarchy of config files and environment variables, as well
    as standard ArgParse command-line parsing behavior.

    We take default values from configuration files:

    - System-wide (see code for location)
    - User-level overrides (see code for location, hopefully under home dir)
    - If a project-specific config file is available, it will override both
      of the above.

    Environment variables will override all configuration files.
    For an option `--foo-bar`, if an environment variable named `YBT_FOO_VAR`
    exists, the option value will be taken from there.

    Of course, options specified directly on the command-line always win.
    """
    global PARSER  # pylint: disable=global-statement
    if PARSER is None:
        config_files = ['/etc/yabt.conf', '~/.yconfig']
        if project_config_file:
            config_files.append(project_config_file)
        PARSER = configargparse.getArgumentParser(
            # Support loading default values from system-wide or
            # user-specific config files (user-level overrides system-wide)
            default_config_files=config_files,
            # Show default values in help message
            formatter_class=configargparse.DefaultsFormatter,
            auto_env_var_prefix='ybt_',
            args_for_setting_config_path=['--config'],
            args_for_writing_out_config_file=['--write-out-config-file'])
        # PARSER.add('--config', is_config_file=True, help='Config file path')
        PARSER.add('--bin-output-dir', default='ybt_bin')
        PARSER.add('--build-file-name', default='YBuild')
        PARSER.add('--build-base-images', action='store_true')
        PARSER.add('--builders-workspace-dir', default='yabtwork')
        PARSER.add('--default-target-name', default='@default')
        PARSER.add('--force-pull', action='store_true')
        # TODO(itamar): support auto-detection of interactivity-mode
        PARSER.add('--non-interactive', action='store_true')
        PARSER.add('--offline', action='store_true')
        # TODO(itamar): this flag should come from the builder, not from here
        PARSER.add('--push', action='store_true')
        PARSER.add('--scm-provider')
        PARSER.add('--docker-volume',
                   help='Use the specified docker volume as buildenv /project')
        PARSER.add('--docker-pull-cmd', default='docker pull',
                   help='Command to use for pulling images from registries')
        PARSER.add('--docker-push-cmd', default='docker push',
                   help='Command to use for pushing images to registries')
        PARSER.add('-v', '--verbose', action='store_true',
                   help='More verbose output to STDOUT')
        PARSER.add('--with-tini-entrypoint', action='store_true')
        PARSER.add('--artifacts-metadata-file',
                   help='Output file to write artifacts metadata to')
        # Logging flags
        PARSER.add('--logtostderr', action='store_true',
                   help='Whether to log to STDERR')
        PARSER.add('--logtostdout', action='store_true',
                   help='Whether to log to STDOUT')
        PARSER.add('--loglevel', default='INFO', choices=LOG_LEVELS_CHOICES,
                   help='Log level threshold')
        PARSER.add('cmd',
                   choices=['build', 'tree', 'version', 'list-builders'],
                   nargs='?', default='build')
        PARSER.add('targets', nargs='*')
    return PARSER


def find_project_config_file(project_root: str) -> str:
    """Return absolute path to project-specific config file, if it exists.

    :param project_root: Absolute path to project root directory.

    A project config file is a file named `YCONFIG_FILE` found at the top
    level of the project root dir.

    Return `None` if project root dir is not specified,
    or if no such file is found.
    """
    if project_root:
        project_config_file = os.path.join(project_root, YCONFIG_FILE)
        if os.path.isfile(project_config_file):
            return project_config_file


def get_user_settings_module(project_root: str):
    """Return project-specific user settings module, if it exists.

    :param project_root: Absolute path to project root directory.

    A project settings file is a file named `YSETTINGS_FILE` found at the top
    level of the project root dir.

    Return `None` if project root dir is not specified,
    or if no such file is found.

    Raise an exception if a file is found, but not importable.

    The YSettings file can define 2 special module-level functions that
    interact with the YABT CLI & config system:
    1. `extend_cli`, if defined, takes the YABT `parser` object and may extend
       it, to add custom command-line flags for the project.
       (careful not to collide with YABT flags...)
    2. `extend_config`, if defined, takes the YABT `config` object and the
       parsed `args` object (returned by the the parser), and may extend the
       config - should be used to reflect custom project CLI flags in the
       config object.

    Beyond that, the settings module is available in YBuild's under
    `conf.settings` (except for the 2 special fucntions that are removed).
    """
    if project_root:
        project_settings_file = os.path.join(project_root, YSETTINGS_FILE)
        if os.path.isfile(project_settings_file):
            settings_loader = SourceFileLoader(
                'settings', project_settings_file)
            return settings_loader.load_module()


def init_and_get_conf(argv: list=None) -> Config:
    """Initialize a YABT CLI environment and return a Config instance.

    :param argv: Manual override of command-line params to parse (for tests).
    """
    colorama.init()
    work_dir = os.path.abspath(os.curdir)
    project_root = search_for_parent_dir(work_dir,
                                         with_files=set([BUILD_PROJ_FILE]))
    parser = make_parser(find_project_config_file(project_root))
    settings_module = get_user_settings_module(project_root)
    if settings_module:
        if hasattr(settings_module, 'extend_cli'):
            settings_module.extend_cli(parser)
    argcomplete.autocomplete(parser)
    args = parser.parse(argv)
    config = Config(args, project_root, work_dir)
    if settings_module:
        if hasattr(settings_module, 'extend_config'):
            settings_module.extend_config(config, args)
        config.settings = settings_module
        try:
            delattr(config.settings, 'extend_cli')
        except AttributeError:
            pass
        try:
            delattr(config.settings, 'extend_config')
        except AttributeError:
            pass
    else:
        config.settings = None
    return config



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt target utils module
~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from collections import defaultdict
from os.path import join, normpath, relpath
import types

from munch import Munch
from ostrich.utils.text import get_safe_path

from .compat import walk
from .config import Config
from .utils import norm_proj_path


_TARGET_NAMES_WHITELIST = frozenset(('*', '@default'))


def validate_name(target_name):
    try:
        if (target_name in _TARGET_NAMES_WHITELIST or
                target_name == get_safe_path(target_name)):
            return target_name
    except ValueError:
        pass
    raise ValueError("Invalid target name: `{}'".format(target_name))


def split(target_name):
    """Split a target name. Returns a tuple "(build_module, name)".

    The split is on the first `:`.
    Extra `:` are considered part of the name.
    """
    return target_name.split(':', 1)


def split_build_module(target_name):
    """Return the build module component of a target name."""
    return split(target_name)[0]


def split_name(target_name):
    """Return the name component of a target name."""
    return split(target_name)[1]


def norm_name(build_module: str, target_name: str):
    """Return a normalized canonical target name for the `target_name`
       observed in build module `build_module`.

    A normalized canonical target name is of the form "<build module>:<name>",
    where <build module> is the relative normalized path from the project root
    to the target build module, and <name> is a valid target name
    (see `validate_name()`).
    """
    if ':' not in target_name:
        raise ValueError(
            "Must provide fully-qualified target name (with `:') to avoid "
            "possible ambiguity - `{}' not valid".format(target_name))
        # return '{}:{}'.format(build_module, validate_name(target_name))

    mod, name = split(target_name)
    return '{}:{}'.format(norm_proj_path(mod, build_module),
                          validate_name(name))


def expand_target_selector(target_selector: str, conf: Config):
    """

    Target specifier can be:

    - `**:*` - means to recursively build all targets under current
      working dir.
    - relative path from current working directory to another directory -
        means to build all targets defined in that build module.
    - a name of a target - means to build this named target in the build module
        in the current working directory.
    - a named target in another build module, with the build module given as a
        relative path from the current working directory (e.g. `../foo:bar`) -
        means to build the specified named target in the specified build
        module.
    - in cases where a relative path can be specified, it should be given using
        standard POSIX relative path construction. in addition, it is possible
        to prefix the path with `@` as a shortcut to the root of the project,
        so when the project root is at `~/foo`, and the working directory is
        `~/foo/bar`, then `@/baz:boom` would refer to a target named "boom" in
        the build module at `~/foo/baz` (just like `../baz:boom`).

    Return a normalized `(build_module_dir, target_name)` tuple.

    # In case if `**:*`, the returned tuple is `('**', '*')`
    #
    # Return a normalized target name (where `**:*` is the normalized form of
    # itself).
    """
    if target_selector == '**:*':
        return target_selector
    if ':' not in target_selector:
        target_selector += ':*'
    build_module, target_name = split(target_selector)
    if build_module.startswith('@'):
        build_module = normpath(relpath(
            build_module.replace('@', conf.project_root, 1),
            conf.project_root))
    else:
        build_module = normpath(join(conf.get_rel_work_dir(), build_module))
    return '{}:{}'.format('' if build_module == '.' else build_module,
                          validate_name(target_name))


def parse_target_selectors(target_selectors: list, conf: Config):
    return [expand_target_selector(target_selector, conf)
            for target_selector in target_selectors]


def generate_build_modules(top: str, conf: Config):
    # TODO(itamar): add ignore marker files / flags
    for root, unused_dirs, files in walk(top):
        if conf.build_file_name in files:
            yield expand_target_selector(root, conf)


class Target(types.SimpleNamespace):  # pylint: disable=too-few-public-methods

    def __init__(self, builder_name):
        super().__init__(name=None, builder_name=builder_name, props=Munch(),
                         deps=None, buildenv=None, tags=set(),
                         artifacts=defaultdict(list))

    def __repr__(self):
        keys = ['name', 'builder_name', 'props', 'deps', 'tags', 'artifacts']
        items = ('{}={!r}'.format(k, self.__dict__[k]) for k in keys)
        return '{}({})'.format(type(self).__name__, ', '.join(items))



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Custom Installer Builder
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from collections import namedtuple
import os
from os.path import basename, isfile, join, relpath, splitext
import shutil
import tarfile
from urllib.parse import urlparse
from zipfile import ZipFile

import git
from git import InvalidGitRepositoryError, NoSuchPathError
import requests

from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..logging import make_logger
from ..target_utils import split_name
from ..utils import yprint


logger = make_logger(__name__)


register_builder_sig(
    'CustomInstaller',
    [('uri', PT.StrList),
     ('script', PT.File),
     ('local_data', PT.FileList, None),
     ('uri_type', PT.StrList, None),
     ('caching', PT.bool, True),
     ])


CustomInstaller = namedtuple('CustomInstaller',
                             ['name', 'package', 'install_script'])


KNOWN_ARCHIVES = frozenset(('.gz', '.bz2', '.tgz', '.zip'))


def guess_uri_type(uri: str, hint: str=None):
    """Return a guess for the URI type based on the URI string `uri`.

    If `hint` is given, it is assumed to be the correct type.
    Otherwise, the URI is inspected using urlparse, and we try to guess
    whether it's a remote Git repository, a remote downloadable archive,
    or a local-only data.
    """
    # TODO(itamar): do this better
    if hint:
        return hint
    norm_uri = uri.lower()
    parsed_uri = urlparse(norm_uri)
    if parsed_uri.path.endswith('.git'):
        return 'git'
    if parsed_uri.scheme in ('http', 'https'):
        ext = splitext(parsed_uri.path)[-1]
        if ext in KNOWN_ARCHIVES:
            return 'archive'
        return 'single'
    return 'local'


def gitfilter(tarinfo):
    """Filter function for tar.add, to filter out git internal stuff."""
    if basename(tarinfo.name) in ['.git', '.gitignore']:
        return None
    return tarinfo


def make_tar(target):
    """Return an open tar object, for writing compressed gz archive."""
    return tarfile.open(target.props.installer_desc.package, 'w:gz',
                        dereference=True)


def git_handler(unused_build_context, target, uri, package_dir, tar):
    """Handle remote Git repository URI.

    Clone the repository under the private builder workspace (unless already
    cloned), and add it to the package tar (filtering out git internals).

    TODO(itamar): Support branches / tags / specific commit hashes
    TODO(itamar): Support updating a cloned repository
    TODO(itamar): Handle submodules?
    TODO(itamar): Handle force pulls?
    """
    target_name = split_name(target.name)
    # clone the repository under a private builder workspace
    try:
        repo = git.Repo(package_dir)
    except (InvalidGitRepositoryError, NoSuchPathError):
        repo = git.Repo.clone_from(uri, package_dir)
    assert repo.working_tree_dir == package_dir
    tar.add(package_dir, arcname=target_name, filter=gitfilter)


def fetch_url(url, dest, parent_to_remove_before_fetch):
    """Helper function to fetch a file from a URL."""
    if isfile(dest):
        logger.debug('File {} is cached', dest)
    else:
        logger.debug('Downloading file {} from {}', dest, url)
        try:
            shutil.rmtree(parent_to_remove_before_fetch)
        except FileNotFoundError:
            pass
        os.makedirs(parent_to_remove_before_fetch)
        # TODO(itamar): Better downloading (multi-process-multi-threaded?)
        # Consider offloading this to a "standalone app" invoked with Docker
        resp = requests.get(url, stream=True)
        with open(dest, 'wb') as fetch_file:
            for chunk in resp.iter_content(chunk_size=32 * 1024):
                fetch_file.write(chunk)


def archive_handler(unused_build_context, target, uri, package_dir, tar):
    """Handle remote downloadable archive URI.

    Download the archive and cache it under the private builer workspace
    (unless already downloaded), extract it, and add the content to the
    package tar.

    TODO(itamar): Support re-downloading if remote changed compared to local.
    TODO(itamar): Support more archive formats (currently only tarballs).
    """
    package_dest = join(package_dir, basename(urlparse(uri).path))
    package_content_dir = join(package_dir, 'content')
    fetch_url(uri, package_dest, package_dir)

    # TODO(itamar): Avoid repetition of splitting extension here and above
    # TODO(itamar): Don't use `extractall` on potentially untrsuted archives
    ext = splitext(package_dest)[-1].lower()
    if ext in ('.gz', '.bz2', '.tgz'):
        with tarfile.open(package_dest, 'r:*') as src_tar:
            src_tar.extractall(package_content_dir)
    elif ext in ('.zip',):
        with ZipFile(package_dest, 'r') as zipf:
            zipf.extractall(package_content_dir)
    else:
        raise ValueError('Unsupported extension {}'.format(ext))
    tar.add(package_content_dir, arcname=split_name(target.name))


def fetch_file_handler(unused_build_context, target, uri, package_dir, tar):
    """Handle remote downloadable file URI.

    Download the file and cache it under the private builer workspace
    (unless already downloaded), and add it to the package tar.

    TODO(itamar): Support re-downloading if remote changed compared to local.
    """
    fetch_url(
        uri, join(package_dir, basename(urlparse(uri).path)), package_dir)
    tar.add(package_dir, arcname=split_name(target.name))


def local_handler(build_context, target, uri, package_dir, tar):
    pass


@register_build_func('CustomInstaller')
def custom_installer_builder(build_context, target):
    # TODO(itamar): Handle cached package invalidation
    yprint(build_context.conf,
           'Fetch and cache custom installer package', target)
    workspace_dir = build_context.get_workspace('CustomInstaller', target.name)
    target_name = split_name(target.name)
    script_name = basename(target.props.script)
    package_tarball = '{}.tar.gz'.format(join(workspace_dir, target_name))
    target.props.installer_desc = CustomInstaller(
        name=target_name, package=package_tarball, install_script=script_name)
    if isfile(package_tarball):
        if target.props.caching:
            logger.debug('Custom installer package {} is cached',
                         package_tarball)
            return
        logger.info('Removing cached custom installer workspace {}',
                    workspace_dir)
        shutil.rmtree(workspace_dir)
        os.makedirs(workspace_dir)

    logger.debug('Making custom installer package {}', package_tarball)
    handlers = {
        'git': git_handler,
        'archive': archive_handler,
        'single': fetch_file_handler,
        'local': local_handler,
    }
    uris = target.props.uri
    uri_types = target.props.uri_type
    # for multiple URI's:
    # 1. either uri_type is empty - so "guess" for all
    # 2. or uri_type has one value - use same value for all URI's
    # 3. or uri_type has multiple values - must be of same length!
    if len(uri_types) == 0 and len(uris) > 0:
        uri_types = [None]
    if len(uris) > 1 and len(uri_types) == 1:
        uri_types = [uri_types[0]] * len(uris)
    assert len(uris) == len(uri_types)
    tar = make_tar(target)
    for i, (uri, uri_type) in enumerate(zip(uris, uri_types)):
        uri_type = guess_uri_type(uri, uri_type)
        logger.debug('CustomInstaller URI {} typed guessed to be {}',
                     uri, uri_type)
        package_dir = join(workspace_dir, target_name,
                           '{}-{}'.format(uri_type, i))
        handlers[uri_type](build_context, target, uri, package_dir, tar)
    # Add local data to installer package, if specified
    for local_node in target.props.local_data:
        tar.add(join(build_context.conf.project_root, local_node),
                arcname=join(target_name, local_node))
    # Add the install script to the installer package
    tar.add(join(build_context.conf.project_root, target.props.script),
            arcname=join(target_name,
                         target.props.installer_desc.install_script))
    tar.close()


@register_manipulate_target_hook('CustomInstaller')
def custom_installer_manipulate_target(build_context, target):
    target.tags.add('custom-installer')



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt target graph tests
~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import networkx
import pytest

from .buildcontext import BuildContext
from .graph import populate_targets_graph, topological_sort
from .extend import Plugin


@pytest.mark.usefixtures('in_dag_project')
def test_target_graph(basic_conf):
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    assert (
        set(['yapi/server:users', ':flask', ':gunicorn', 'common:logging',
             'fe:fe', 'yapi/server:yapi', 'yapi/server:yapi-gunicorn',
             'common:base']) == set(build_context.target_graph.nodes()))
    assert (
        set([('fe:fe', 'yapi/server:users'), ('fe:fe', ':flask'),
             ('fe:fe', 'common:base'), ('yapi/server:yapi', ':flask'),
             ('yapi/server:yapi', 'common:base'),
             ('yapi/server:yapi-gunicorn', 'yapi/server:yapi'),
             ('yapi/server:yapi-gunicorn', 'common:base'),
             ('yapi/server:yapi-gunicorn', ':gunicorn'),
             ('common:base', 'common:logging')]) ==
        set(build_context.target_graph.edges()))
    # Can't assert the list directly, because it is not stable / deterministic
    topo_sort = list(topological_sort(build_context.target_graph))

    def assert_dep_chain(*chain):
        dep_chain = [topo_sort.index(target_name) for target_name in chain]
        assert dep_chain == sorted(dep_chain)

    assert_dep_chain('common:logging', 'common:base',
                     'yapi/server:yapi', 'yapi/server:yapi-gunicorn')
    assert_dep_chain(':flask', 'yapi/server:yapi')
    assert_dep_chain(':gunicorn', 'yapi/server:yapi-gunicorn')
    assert_dep_chain(':flask', 'fe:fe')
    assert_dep_chain('yapi/server:users', 'fe:fe')
    assert_dep_chain('common:logging', 'common:base', 'fe:fe')


def test_stable_topological_sort():
    """Test that my modified topological sort is stable.

    Note - not doing many cycles because I saw that even with the non-stable
    implementation, it is stable in the context of the same process...
    (only reruns showed the unstable behavior)
    """
    expected_order = ['world', 'bar', 'baz', 'hello', 'foo']

    graph = networkx.DiGraph()
    graph.add_edges_from([('foo', 'bar'), ('foo', 'baz'),
                          ('bar', 'world'), ('foo', 'hello')])
    assert list(topological_sort(graph)) == expected_order

    same_graph = networkx.DiGraph({'foo': ['baz', 'hello', 'bar']})
    same_graph.add_edge('bar', 'world')
    assert list(topological_sort(same_graph)) == expected_order


def test_topological_sort1():
    graph = networkx.DiGraph()
    graph.add_edges_from([(1, 2), (1, 3), (2, 3)])
    assert list(topological_sort(graph)) == [3, 2, 1]

    graph.add_edge(3, 2)
    with pytest.raises(networkx.NetworkXUnfeasible):
        list(topological_sort(graph))

    graph.remove_edge(2, 3)
    assert list(topological_sort(graph)) == [2, 3, 1]


def test_topological_sort2():
    graph = networkx.DiGraph({1: [2], 2: [3], 3: [4],
                              4: [5], 5: [1], 11: [12],
                              12: [13], 13: [14], 14: [15]})

    with pytest.raises(networkx.NetworkXUnfeasible):
        list(topological_sort(graph))

    graph.remove_edge(1, 2)
    assert list(topological_sort(graph)) == [1, 5, 4, 3, 2, 15, 14, 13, 12, 11]


def test_topological_sort4():
    graph = networkx.Graph()
    graph.add_edge(1, 2)
    with pytest.raises(networkx.NetworkXError):
        list(topological_sort(graph))


def test_topological_sort4():
    graph = networkx.DiGraph()
    graph.add_edge(1, 2)
    assert list(topological_sort(graph)) == [2, 1]



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt ProtoBuf builder
~~~~~~~~~~~~~~~~~~~~~

:author: Zohar Rimon, Itamar Ostricher
"""


import os
from os.path import dirname, isfile, join, relpath, splitext
from pathlib import Path

from ostrich.utils.path import commonpath

from ..compat import walk
from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..logging import make_logger
from ..utils import link_artifacts, yprint


logger = make_logger(__name__)


register_builder_sig(
    'Proto',
    [('sources', PT.FileList),
     ('in_buildenv', PT.Target),
     ('proto_cmd', PT.list, 'protoc'),
     ('gen_python', PT.bool, True),
     ('gen_cpp', PT.bool, True),
     ('gen_python_rpcz', PT.bool, False),
     ('gen_cpp_rpcz', PT.bool, False),
     ('gen_python_grpc', PT.bool, False),
     ('gen_cpp_grpc', PT.bool, False),
     ('grpc_plugin_path', PT.str, '/usr/local/bin/grpc_cpp_plugin'),
     ('copy_generated_to', PT.File, None),
     ('cmd_env', None),
     ])


@register_build_func('Proto')
def proto_builder(build_context, target):
    yprint(build_context.conf, 'Build ProtoBuf', target)
    workspace_dir = build_context.get_workspace('ProtoBuilder', target.name)
    proto_dir = join(workspace_dir, 'proto')
    # Collect proto sources from this target and all dependecies,
    # to link under the target sandbox dir
    protos = [source for source in target.props.sources
              if source.endswith('.proto')]
    for dep in build_context.walk_target_deps_topological_order(target):
        if 'sources' in dep.props:
            protos.extend(source for source in dep.props.sources
                          if source.endswith('.proto'))
    link_artifacts(protos, proto_dir, None, build_context.conf)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_dir)
    protoc_cmd = target.props.proto_cmd + ['--proto_path', buildenv_workspace]
    if target.props.gen_cpp:
        protoc_cmd.extend(('--cpp_out', buildenv_workspace))
    if target.props.gen_python:
        protoc_cmd.extend(('--python_out', buildenv_workspace))
    if target.props.gen_cpp_grpc:
        protoc_cmd.extend(
            ('--grpc_out', buildenv_workspace,
             '--plugin=protoc-gen-grpc={}'
             .format(target.props.grpc_plugin_path)))
    if target.props.gen_python_grpc:
        protoc_cmd.extend(('--grpc_python_out', buildenv_workspace))
    if target.props.gen_cpp_rpcz:
        protoc_cmd.extend(('--cpp_rpcz_out', buildenv_workspace))
    if target.props.gen_python_rpcz:
        protoc_cmd.extend(('--python_rpcz_out', buildenv_workspace))
    protoc_cmd.extend(join(buildenv_workspace, 'proto', src)
                      for src in target.props.sources)
    build_context.run_in_buildenv(
        target.props.in_buildenv, protoc_cmd, target.props.cmd_env)
    generated_files = []
    target.artifacts['gen'] = {}

    def add_artifact(file_path):
        target.artifacts['gen'][relpath(file_path, workspace_dir)] = (
            relpath(file_path, build_context.conf.project_root))

    def process_generated(src_base, gen_suffixes, is_artifact):
        gen_files = [src_base + gen_suffix for gen_suffix in gen_suffixes]
        if not all(isfile(gen_path) for gen_path in gen_files):
            logger.error('Missing expected generated files: {}',
                         ', '.join(gen_files))
        else:
            if is_artifact:
                for gen_file in gen_files:
                    add_artifact(gen_file)
            generated_files.extend(gen_files)

    def create_init_py(path: str):
        init_py_path = join(path, '__init__.py')
        Path(init_py_path).touch(exist_ok=True)
        return init_py_path

    # Add generated files to artifacts / generated list
    for src in target.props.sources:
        src_base = join(proto_dir, splitext(src)[0])
        if target.props.gen_python:
            process_generated(src_base, ('_pb2.py',), True)
        if target.props.gen_cpp:
            process_generated(src_base, ('.pb.cc', '.pb.h'), False)
        if target.props.gen_python_rpcz:
            process_generated(src_base, ('_rpcz.py',), True)
        if target.props.gen_cpp_rpcz:
            process_generated(src_base, ('.rpcz.cc', '.rpcz.h'), False)
        if target.props.gen_python_grpc:
            process_generated(src_base, ('_pb2_grpc.py',), True)
        if target.props.gen_cpp_grpc:
            process_generated(src_base, ('.grpc.pb.cc', '.grpc.pb.h'), False)

    # Create __init__.py files in all generated directories with Python files
    if target.props.gen_python or target.props.gen_python_rpcz:
        join_env = target.props.packaging_params.pop('semicolon_join_env', {})
        if 'PYTHONPATH' in join_env:
            if '/usr/src/gen' not in join_env['PYTHONPATH'].split(':'):
                join_env['PYTHONPATH'] += ':/usr/src/gen'
        else:
            join_env['PYTHONPATH'] = '/usr/src/gen'
        target.props.packaging_params['semicolon_join_env'] = join_env
        py_dirs = set(('',))
        for src in target.props.sources:
            py_dir = dirname(src)
            while py_dir not in py_dirs:
                py_dirs.add(py_dir)
                py_dir = dirname(py_dir)
        for py_dir in py_dirs:
            add_artifact(create_init_py(join(proto_dir, py_dir)))

    # Copy generated files to external destination
    if target.props.copy_generated_to:
        link_artifacts(generated_files, target.props.copy_generated_to,
                       workspace_dir, build_context.conf)
        if target.props.gen_python or target.props.gen_python_rpcz:
            # Create __init__.py files in external destination dirs too
            for root, dirs, unused_files in walk(
                    target.props.copy_generated_to):
                for proto_dir in dirs:
                    create_init_py(join(root, proto_dir))


@register_manipulate_target_hook('Proto')
def proto_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv



# -*- coding: utf-8 -*-

# Copyright 2016 Yowza Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Apt builder tests
~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import pytest

from . import apt
from ..target_utils import Target


slow = pytest.mark.skipif(not pytest.config.getoption('--with-slow'),
                          reason='need --with-slow option to run')


DISTRO = {
    'id': 'Ubuntu',
    'release': '14.04',
    'codename': 'trusty',
    'description': 'Ubuntu 14.04.4 LTS',
}


def test_apt_repository_simple_line():
    target = Target('AptRepository')
    llvm_repo = ('deb http://llvm.org/apt/trusty/ '
                 'llvm-toolchain-trusty-3.8 main')
    target.props.source = llvm_repo
    target.props.key = None
    target.props.keyserver = 'hkp://keyserver.ubuntu.com:80'
    source_line, apt_key_cmd = apt.parse_apt_repository(None, target, DISTRO)
    assert llvm_repo == source_line
    assert apt_key_cmd is None


@slow
def test_apt_repository_ppa_with_key():
    target = Target('AptRepository')
    target.props.source = 'ppa:brightbox/ruby-ng'
    target.props.key = None
    target.props.keyserver = 'hkp://keyserver.ubuntu.com:80'
    source_line, apt_key_cmd = apt.parse_apt_repository(None, target, DISTRO)
    exp_ruby_ng_repo = ('deb http://ppa.launchpad.net/brightbox/ruby-ng/'
                        'ubuntu trusty main')
    exp_ruby_ng_key_cmd = (
        'apt-key adv --keyserver hkp://keyserver.ubuntu.com:80 '
        '--recv 80F70E11F0F0D5F10CB20E62F5DA5F09C3173AA6')
    assert exp_ruby_ng_repo == source_line
    assert exp_ruby_ng_key_cmd == apt_key_cmd
